<?php

namespace UW\FCS\Entity;

use XF\BbCode\RenderableContentInterface;
use XF\Entity\Attachment;
use XF\Entity\LinkableInterface;
use XF\Mvc\Entity\Structure;
use XF\Mvc\Entity\Entity;
use XF\Entity\ReactionTrait;

class Comment extends \XF\Mvc\Entity\Entity implements \XF\Entity\QuotableInterface , RenderableContentInterface,LinkableInterface
{
    use ReactionTrait;
    
    public $mentionedUsers = [];
    
    public function getQuoteWrapper($inner)
    {
            return '[QUOTE="'
                    . ($this->User ? $this->User->username : $this->username)
                    . ', comment: ' . $this->comment_id
                    . ($this->User ? ", member: $this->user_id" : '')
                    . '"]'
                    . "\n" . $inner . "\n"
                    . "[/QUOTE]\n";
    }
    
    /**
     * lets create a new structure
     * @param Structure $structure
     * @return Structure
     */
    public static function getStructure(Structure $structure)
    {                
        $structure->table = 'xf_uw_comment';
        $structure->shortName = 'UW\FCS:Comment';
        $structure->contentType = 'post_comment';        
        $structure->primaryKey = 'comment_id';
        $structure->columns = [
            'comment_id' => ['type' => self::UINT, 'autoIncrement' => true, 'nullable' => true],
            'message' => ['type' => self::STR,
				'required' => 'please_enter_valid_message'
			],
            'post_id' => ['type' => self::UINT, 'required' => true],
            'thread_id' => ['type' => self::UINT, 'required' => true],
            'user_id' => ['type' => self::UINT, 'required' => true],
            'comment_date' => ['type' => self::INT, 'default' => \XF::$time],
            'last_edit_date' => ['type' => self::INT, 'default' => 0],
            'edit_count' => ['type' => self::INT, 'default' => 0],
            'status' => ['type' => self::INT, 'default' => 1],
            'warning_id' => ['type' => self::UINT, 'default' => 0],
            'warning_message' => ['type' => self::STR, 'default' => '', 'api' => true],            
            'ip_id' => ['type' => self::UINT, 'default' => 0],
            'attach_count' => ['type' => self::UINT, 'max' => 65535, 'forced' => true, 'default' => 0],
            'username' => ['type' => self::STR, 'maxLength' => 50,
				'required' => 'please_enter_valid_name', 
			],
            'embed_metadata' => ['type' => self::JSON_ARRAY, 'nullable' => true, 'default' => null]
        ];
              
//        $structure->getters['username'] = ['getter' => 'getUsername', 'cache' => false];
        $structure->getters['post_date'] = ['getter' => 'getPostDate', 'cache' => false];
        $structure->getters = [
            'Unfurls' => true
        ];
        $structure->relations = [    
        'User' => [
				'entity' => 'XF:User',
				'type' => self::TO_ONE,
				'conditions' => 'user_id',
				'primary' => true
			],
        'Post' => [
				'entity' => 'XF:Post',
				'type' => self::TO_ONE,
				'conditions' => 'post_id',
				'primary' => true
			],
        'Thread' => [
				'entity' => 'XF:Thread',
				'type' => self::TO_ONE,
				'conditions' => 'thread_id',
				'primary' => true
			],
        'DeletionLog' => [
                            'entity' => 'XF:DeletionLog',
                            'type' => self::TO_ONE,
                            'conditions' => [
                                    ['content_type', '=', 'comment'],
                                    ['content_id', '=', '$comment_id']
                            ],
                            'primary' => true
                    ],
        'ApprovalQueue' => [
                            'entity' => 'XF:ApprovalQueue',
                            'type' => self::TO_ONE,
                            'conditions' => [
                                    ['content_type', '=', 'post_comment'],
                                    ['content_id', '=', '$comment_id']
                            ],
                            'primary' => true
                    ],
            'Attachments' => [
                'entity' => 'XF:Attachment',
                'type' => self::TO_MANY,
                'conditions' => [
                    ['content_type', '=', 'post_comment'],
                    ['content_id', '=', '$comment_id']
                ],
                'with' => 'Data',
                'order' => 'attach_date'
            ]
        ];
        
        $structure->behaviors = [			
			'XF:NewsFeedPublishable' => [
				'usernameField' => 'username',
				'dateField' => 'comment_date'
			],
			'XF:Indexable' => [
				'checkForUpdates' => ['message', 'user_id', 'comment_date', 'status']
			],
                        'XF:Reactable' => ['stateField' => 'status'],
                    ];
        
        static::addReactableStructureElements($structure);
        return $structure;
    }
    
    public function getUsername()
    {
        if($this->user_id)
        {
            return $this->User->username;
        }
        return $this->username;
    }
    public function getPostDate()
    {
        return $this->comment_date;
    }
    
    public function canEditSilently()
    {
        $visitor = \XF::visitor();                
        $nodeId = $this->Thread->node_id;                
                
        if ($visitor->hasNodePermission($nodeId, 'uw_fcs_edit_comment_silen'))
        {
            return TRUE;
        }
        return FALSE;
    }
    public function canViewHistory()
    {
        $visitor = \XF::visitor();
        if (!$visitor->user_id)
        {
                return false;
        }

        if (!$this->app()->options()->editHistory['enabled'])
        {
                return false;
        }
        
        if ($visitor->hasNodePermission($this->Thread->node_id, 'uw_fcs_edit_comment'))
        {
            return TRUE;
        }  
        return FALSE;
    }

    public function softDelete($reason = '', User $byUser = null)
    {
            $byUser = $byUser ?: \XF::visitor();          

            if ($this->status == 2)
            {
                    return false;
            }

            $this->status = 2;

            /** @var \XF\Entity\DeletionLog $deletionLog */
            $deletionLog = $this->getRelationOrDefault('DeletionLog');
            $deletionLog->setFromUser($byUser);
            $deletionLog->delete_reason = $reason;
            $this->save();
            return true;
    }
        
    public function canEdit()
    {        
        $visitor = \XF::visitor();
        if ($visitor->hasNodePermission($this->Thread->node_id, 'uw_fcs_edit_comment'))
        {
            return TRUE;
        }
        if($visitor->hasNodePermission($this->Thread->node_id, 'uw_fcs_edit_comment_own') && ($this->user_id == $visitor->user_id))
        {
            $editLimit = $visitor->hasNodePermission($this->Thread->node_id, 'uw_fcs_edit_comment_time');
                        
            if ($editLimit != -1 && (!$editLimit || $this->comment_date < \XF::$time - 60 * $editLimit))
            {
                return false;
            }            
            return TRUE;
        }
        return FALSE;
    }
    
    public function isUnread()
    {
        if (!$this->Thread)
        {
                return false;
        }

        $readDate = $this->Thread->getVisitorReadDate();
        if ($readDate === null)
        {
                return false;
        }

        return $readDate < $this->comment_date;
    }
    
    public function canDelete()
    {
        $visitor = \XF::visitor();
        if ($visitor->hasNodePermission($this->Thread->node_id, 'uw_fcs_delete_comment'))
        {
            return TRUE;
        }
                 
        if($visitor->hasNodePermission($this->Thread->node_id, 'uw_fcs_delete_comment_own') && ($this->user_id == $visitor->user_id))
        {
            $editLimit = $visitor->hasNodePermission($this->Thread->node_id, 'uw_fcs_edit_comment_time');
                        
            if ($editLimit != -1 && (!$editLimit || $this->comment_date < \XF::$time - 60 * $editLimit))
            {
                return false;
            }            
            return TRUE;
        }        
        return FALSE;
    }        

    protected function _postDelete()
    {
        if(\XF::options()->uw_count_comment_as_reply)
        {
            $this->Post->Thread->reply_count--;
            $this->Post->Thread->save();
            if($this->User && $this->User->message_count > 0)
            {
                $this->User->fastUpdate('message_count', ($this->User->message_count - 1));
            }
        }
        $attachRepo = $this->repository('XF:Attachment');
        $attachRepo->fastDeleteContentAttachments('post_comment', $this->comment_id);
        return parent::_postDelete();
    }
    
    protected function _postSave()
    {
        $mentionedUsersIds = array_keys($this->mentionedUsers);
        $approvalChange = $this->isStateChanged('status', 3);
                
        $visibilityChange = $this->isStateChanged('status', '1');
        if ($this->isInsert() || ($this->isUpdate() && ($visibilityChange == 'enter' || $visibilityChange == 'leave')))
        {
                $reactionRepo = $this->repository('XF:Reaction');
                $reactionRepo->recalculateReactionIsCounted('post_comment', $this->comment_id);
                if(\XF::options()->uw_count_comment_as_reply)
                {
                    if($visibilityChange == 'enter')
                    {
                        $this->Post->Thread->reply_count++;
                        if($this->User)
                        {
                            $this->User->fastUpdate('message_count', ($this->User->message_count + 1));
                        }
                    }else
                    {
                        if($this->User)
                        {
                            $this->User->fastUpdate('message_count', ($this->User->message_count - 1));
                        }
                        $this->Post->Thread->reply_count--;
                    }
                    $this->Post->Thread->save();
                }
        }
        
        if(($this->isInsert() && $this->status == 1) || ($this->isUpdate() && $approvalChange == 'leave' && $this->status != 2))
        {
            $this->Post->Thread->last_post_date = \XF::$time;
            $this->Post->Thread->last_post_user_id = $this->user_id;
            $this->Post->Thread->last_post_username = $this->username;
            $this->Post->Thread->save();
            
            $notifier = \XF::app()->service('UW\FCS\Service\Comment\Notifier', $this);
            $notifier->setMentionedUserIds($mentionedUsersIds);
//            $notifier->setQuotedUserIds($this->postPreparer->getQuotedUserIds());
            $notifier->notifyAndEnqueue(3);
        }

        $deletionChange = $this->isStateChanged('status', '2');            

        if ($deletionChange == 'enter' && !$this->DeletionLog)
        {
            $delLog = $this->getRelationOrDefault('DeletionLog', false);
            $delLog->setFromVisitor();
            $delLog->save();
        }
        
        if ($this->isUpdate())
        {
                if ($approvalChange == 'leave' && $this->ApprovalQueue)
                {
                        $this->ApprovalQueue->delete();
                }
        }
        
        if ($approvalChange == 'enter')
        {
                $approvalQueue = $this->getRelationOrDefault('ApprovalQueue', false);
                $approvalQueue->content_date = $this->comment_date;
                $approvalQueue->save();
        }
        
        if($this->isInsert() || ($this->isUpdate() && \XF::visitor()->user_id == $this->user_id))
        {
            $ipRepo = \XF::repository('XF:Ip');
            $ip =  \XF::app()->request()->getIp();        
            $ipEnt = $ipRepo->logIp($this->user_id, $ip, 'post_comment', $this->comment_id);
            if ($ipEnt)
            {
                $this->fastUpdate('ip_id', $ipEnt->ip_id);
            }
        }
        
        parent::_postSave();
    }

    public function getBbCodeRenderOptions($context, $type)
    {
        return [
            'entity' => $this,
            'user' => $this->User,
            'treatAsStructuredText' => true,
            'attachments' => $this->attach_count ? $this->Attachments : [],
            'viewAttachments' => $this->canViewAttachments(),
            'unfurls' => $this->Unfurls ?: []
        ];
    }

    public function getUnfurls()
    {
        return $this->_getterCache['Unfurls'] ?? [];
    }

    public function setUnfurls($unfurls)
    {
        $this->_getterCache['Unfurls'] = $unfurls;
    }
    public function canViewAttachments(&$error = null): bool
    {
        $profilePost = $this->Thread;
        if (!$profilePost)
        {
            return false;
        }

        return $profilePost->canViewAttachments($error);
    }

    public function isAttachmentEmbedded($attachmentId)
    {
        if (!$this->embed_metadata)
        {
            return false;
        }

        if ($attachmentId instanceof Attachment)
        {
            $attachmentId = $attachmentId->attachment_id;
        }

        return isset($this->embed_metadata['attachments'][$attachmentId]);
    }

    /**
     * react
     * @param type $eraror
     * @return boolean
     */
    public function canReact(&$error = null)
    {
        $visitor = \XF::visitor();
        if (!$visitor->user_id)
        {
                return false;
        }

        if ($this->status != 1)
        {
                return false;
        }

        if ($this->user_id == $visitor->user_id)
        {
                $error = \XF::phraseDeferred('reacting_to_your_own_content_is_considered_cheating');
                return false;
        }

        if (!$this->Thread)
        {
                return false;
        }

        return $visitor->hasNodePermission($this->Thread->node_id, 'uw_fcs_comment_react');
    }
    
    
    public function canView(&$error = null)
    {
        $visitor = \XF::visitor();
        
        if (!$this->Thread || !$this->Thread->canView($error))
        {
                return false;
        }
        if (!$this->Post || !$this->Post->canView($error))
        {
                return false;
        }
        
        if ($this->status == 3)
        {
                if (!$this->canViewModeratedComments())
                {
                        $error = \XF::phraseDeferred('requested_comment_not_found');
                        return false;
                }
        }
        else if ($this->status == 2)
        {
                if (!$this->canViewDeletedComments())
                {
                        $error = \XF::phraseDeferred('requested_comment_not_found');
                        return false;
                }
        }

        return true;
    }
    
    public function canViewModeratedComments()
    {
        $visitor = \XF::visitor();
        if ($visitor->hasNodePermission($this->Thread->node_id, 'uw_fcs_view_moderated') 
                || ($visitor->user_id && ($visitor->user_id == $this->user_id)))
        {
            return TRUE;
        }
        return FALSE;
    }
    
    public function canViewDeletedComments()
    {
        if (\XF::visitor()->hasNodePermission($this->Thread->node_id, 'uw_fcs_view_deleted_comme'))
        {
            return TRUE;
        }
        return FALSE;
        
    }
    
    public function canApproveUnapprove(&$error = null)
    {
            $visitor = \XF::visitor();
            return ($visitor->user_id && $visitor->hasNodePermission($this->Thread->node_id, 'uw_fcs_approveUnapprove'));
    }

    public function isVisible()
    {
        return (
            $this->status == 1
            && $this->Thread
            && $this->Post
            && $this->Thread->discussion_state == 'visible'
            && $this->Post->message_state == 'visible'
            );
    }
 
    public function canCleanSpam()
    {
        return (\XF::visitor()->canCleanSpam() && $this->User && $this->User->isPossibleSpammer());
    }
    public function canWarn(&$error = null)
    {
            $visitor = \XF::visitor();

            if ($this->warning_id
                    || !$this->user_id
                    || !$visitor->user_id
                    || $this->user_id == $visitor->user_id
                    || !$visitor->hasNodePermission($this->Thread->node_id, 'warn')
            )
            {
                    return false;
            }

            return ($this->User && $this->User->isWarnable());
    }
    
    public function onWarning(Entity $entity, Warning $warning)
    {
            $entity->warning_id = $warning->warning_id;
            $entity->save();
    }
    
    public function isIgnored()
    {
            return \XF::visitor()->isIgnoring($this->user_id);
    }
    
    public function onWarningRemoval(Entity $entity, Warning $warning)
    {
            $entity->warning_id = 0;
            $entity->warning_message = '';
            $entity->save();
    }
    public function getContentUrl(bool $canonical = false, array $extraParams = [], $hash = null)
    {
        $route = $canonical ? 'canonical:comments' : 'comments';
        return $this->app()->router('public')->buildLink($route, $this, $extraParams, $hash);
    }

    public function getContentPublicRoute()
    {
        return 'comments';
    }

    public function getContentTitle(string $context = '')
    {
        return \XF::phrase('uw_fcs_post_comment_by_x', ['title' => $this->username]);
    }
}