<?php

namespace UW\FCS\Notifier\Comment;

use XF\Notifier\AbstractNotifier;

class ThreadWatch extends AbstractNotifier
{
    
    /**
	 * @var \XF\Entity\Post
	 */
	protected $comment;

	protected $actionType;

	protected $userReadDates = [];
	protected $previousComments = null;

	public function __construct(\XF\App $app, \UW\FCS\Entity\Comment $comment, $actionType)
	{
		parent::__construct($app);

		$this->comment = $comment;
		$this->actionType = $actionType;
	}

	public function canNotify(\XF\Entity\User $user)
	{
		if (!isset($this->userReadDates[$user->user_id]))
		{
			// this should have a record for every user, so generally shouldn't happen
			return false;
		}

		$userReadDate = $this->userReadDates[$user->user_id];
		$comment = $this->comment;

		if ($user->user_id == $comment->user_id || $user->isIgnoring($comment->user_id))
		{
			return false;
		}
                if ($userReadDate > $comment->Post->Thread->last_post_date)
		{
			return false;
		}
                
                $previousVisiblePost = $this->getPreviousComments($user);

                $autoReadDate = \XF::$time - $this->app()->options()->readMarkingDataLifetime * 86400;
                if (!$previousVisiblePost || !$previousVisiblePost->comment_id || $previousVisiblePost->comment_date < $autoReadDate)
                {
                        // always alert
                }
                else if ($previousVisiblePost->comment_date > $userReadDate)
                {
                        return false;
                }                
		return true;
	}

	public function sendAlert(\XF\Entity\User $user)
	{
		$comment = $this->comment;
                $extra['also'] = 0;
                $action = 'insert';
                if($user->user_id == $comment->Post->user_id)
                {
                    $extra['also'] = 1;
                    $action = 'insert_my';
                }
		return $this->basicAlert($user, $comment->user_id, $comment->username, 'post_comment', $comment->comment_id, $action,$extra);
	}

	public function sendEmail(\XF\Entity\User $user)
	{
		if (!$user->email || $user->user_state != 'valid')
		{
			return false;
		}

		$comment = $this->comment;

                $also = 0;
                $action = 'post_comment_insert';
                if($user->user_id == $comment->Post->user_id)
                {
                    $also = 1;
                    $action = 'post_comment_insert_my';
                }
                
                $params = [
                    'comment' => $comment,
                    'receiver' => $user,
                    'sender' => $comment->User,
                    'also' => $also
                ];
        
                $this->app()->mailer()->newMail()
                    ->setToUser($user)
                    ->setTemplate($action, $params)
                    ->queue();
                
		return true;
	}

	public function getDefaultNotifyData()
	{
		return $this->getDefaultWatchNotifyData();
	}

	public function getUserData(array $userIds)
	{
		$users = parent::getUserData($userIds);
		$this->userReadDates = $this->getUserReadDates($userIds);

		return $users;
	}

	protected function getUserReadDates(array $userIds)
	{
		if (!$userIds)
		{
			return [];
		}

		$autoReadDate = \XF::$time - $this->app()->options()->readMarkingDataLifetime * 86400;
		$comment = $this->comment;

		$db = $this->app()->db();
		$readDates = $db->fetchPairs("
			SELECT user.user_id,
				GREATEST(
					COALESCE(thread_read.thread_read_date, 0),
					COALESCE(forum_read.forum_read_date, 0),
					?
				)
			FROM xf_user AS user
			LEFT JOIN xf_thread_read AS thread_read ON
				(thread_read.user_id = user.user_id AND thread_read.thread_id = ?)
			LEFT JOIN xf_forum_read AS forum_read ON
				(forum_read.user_id = user.user_id AND forum_read.node_id = ?)
			WHERE user.user_id IN (" . $db->quote($userIds) . ")
		", [$autoReadDate, $comment->Post->thread_id, $comment->Post->Thread->node_id]);

		foreach ($userIds AS $userId)
		{
			if (!isset($readDates[$userId]))
			{
				$readDates[$userId] = $autoReadDate;
			}
		}

		return $readDates;
	}

	protected function getPreviousComments($user)
	{
		if ($this->previousComments === null)
		{
			$autoReadDate = \XF::$time - $this->app()->options()->readMarkingDataLifetime * 86400;

			$finder = $this->app()->finder('UW\FCS:Comment')
				->where('status', 1)
				->where('comment_date', '<', $this->comment->comment_date)
				->where('comment_date', '>=', $autoReadDate)
				->order('comment_date', 'desc');

                        if($user->Option->comment_notif_all)
                        {
                            $finder->where('thread_id', $this->comment->thread_id);
                        }
                        else
                        {
                            $finder->where('post_id', $this->comment->post_id);
                        }
                        
			$this->previousComments = $finder->fetchOne();
		}

		return $this->previousComments;
	}	

	public function getDefaultWatchNotifyData()
	{
		$comment = $this->comment;
                
		$finder = $this->app()->finder('XF:ThreadWatch');

		$finder->where('thread_id', $comment->Post->thread_id)
			->where('User.user_state', '=', 'valid')
			->where('User.is_banned', '=', 0);

		$activeLimit = $this->app()->options()->watchAlertActiveOnly;
		if (!empty($activeLimit['enabled']))
		{
			$finder->where('User.last_activity', '>=', \XF::$time - 86400 * $activeLimit['days']);
		}

		$notifyData = [];
		foreach ($finder->fetchColumns(['user_id', 'email_subscribe']) AS $watch)
		{
			$notifyData[$watch['user_id']] = [
				'alert' => true,
				'email' => (bool)$watch['email_subscribe']
			];
		}

		return $notifyData;
	}

	protected function getWatchEmailTemplateName()
	{
		return 'watched_thread_reply';
	}
}